/*
     eibdump.c - EIBDump Busmonitor for low level EIB driver (BCU 1)

     Copyright (C) 2000 Bernd Thallner <bernd@kangaroo.at>

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "eib.h"
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>

//#include <sys/time.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#define EIB_DEVICE "/dev/eib"

static char *eibdump_version="0.2";

int eib_fd, bcu_version, no_write=0;
char *argv0, *eib_device;

void error(void) {
  char str[64];

  snprintf(str, 64, "%s: %s", argv0, eib_device);
  perror(str);
  close(eib_fd);
  exit(1);
}

void open_eib(void) {

  if(no_write) {
    if((eib_fd=open(eib_device, O_RDONLY))==-1)
      error();
  } else
    if((eib_fd=open(eib_device, O_RDWR))==-1)
      error();


  if((bcu_version=ioctl(eib_fd, EIB_GET_BCU_VERSION))==-1)
    error();

  if(bcu_version!=1)
    fprintf(stderr, "%s: %s: warning unsupported bcu version\n", argv0, eib_device);
}

void close_eib() {


  if(close(eib_fd)==-1) {
  fprintf(stderr, "aaa=%d\n", eib_fd);    
    error();
  }
}

void device_information(void) {
  struct eib_device_information eib_dev_info;

  open_eib();
  if(ioctl(eib_fd, EIB_GET_DEVICE_INFORMATION, &eib_dev_info)==-1)
    error();
  close_eib();
  fprintf(stdout, "%s\n", eib_dev_info.string);
  fprintf(stdout, "warning_flag=0x%08lx\n", eib_dev_info.warning_flag);
  fprintf(stdout, "rb_length=%d\n", eib_dev_info.rb_length);
  fprintf(stdout, "wb_length=%d\n", eib_dev_info.wb_length);
  fprintf(stdout, "rb_max_length=%d\n", eib_dev_info.rb_max_length);
  fprintf(stdout, "wb_max_length=%d\n", eib_dev_info.wb_max_length);
  fprintf(stdout, "stat_read_bytes=%ld\n", eib_dev_info.stat_rx_bytes);
  fprintf(stdout, "stat_write_bytes=%ld\n", eib_dev_info.stat_tx_bytes);
  fprintf(stdout, "stat_read_data_blocks=%ld\n", eib_dev_info.stat_rx_data_blocks);
  fprintf(stdout, "stat_write_data_blocks=%ld\n", eib_dev_info.stat_tx_data_blocks);
  fprintf(stdout, "stat_max_read_time_data_block=%ld\n", eib_dev_info.stat_max_rx_time_data_block);
  fprintf(stdout, "stat_max_write_time_data_block=%ld\n", eib_dev_info.stat_max_tx_time_data_block);
}

int strtoeib(char *in, char *out) {
  int state=0, len=0, tmp;
  do {
    if(state<2) {
      if(*in>='0' && *in<='9')
	tmp=*in-'0';
      else if(*in>='a' && *in<='f')
	tmp=*in-'a'+10;
      else if(*in>='A' && *in<='F')
	tmp=*in-'A'+10;
      else
	return 0;
      if(state)
	out[len]<<=4;
      else
	out[len]=0;
      out[len]+=tmp;
      if(state)
	len++;
      state++;
    } else {
      if(*in!=' ')
	return 0;
      state=0;
    }
  } while(*(++in));
  if(state!=2)
    return 0;
  return len;
}

void write_data_block(char *str) {
  int len;
  char buf[64];

  open_eib();
  if((len=strtoeib(str, buf))==0) {
    fprintf(stderr, "%s: Syntax error in string\n", argv0);
    close_eib();
    exit(1);
  }
  if(write(eib_fd, buf, len)==-1)
    error();
  fprintf(stdout, "write data block: %s\n", str);
  close_eib();
}

void clear_write_buffer(void) {

  open_eib();
  if(ioctl(eib_fd, EIB_CLEAR_WRITE_BUFFER)==-1)
    error();
  fprintf(stdout, "write buffer cleared\n");
  close_eib();
}

void reset_device_statistic(void) {

  open_eib();
  if(ioctl(eib_fd, EIB_RESET_DEVICE_STATISTIC)==-1)
    error();
  fprintf(stdout, "device statistic reseted\n");
  close_eib();
}

void reset_bcu(void) {

  open_eib();
  if(write(eib_fd, "\x46\x01\x00\x60\xc0", 5)==-1) // Busmonitor
    error();
  fprintf(stdout, "bcu reseted\n");
  close_eib();
}

void dump(char *str) {
  time_t t;
  struct tm *local_time;

  time(&t);
  local_time=localtime(&t);
  fprintf(stdout, "%02d:%02d:%02d  %s", local_time->tm_hour, local_time->tm_min, local_time->tm_sec, str);
}

void cleanup(int signal)
{
  if(!no_write) {
    if(write(eib_fd, "\x46\x01\x00\x60\xc0", 5)==-1) // Busmonitor
      error();
    dump("reset bcu\n");
  }
  close_eib();
  exit(1);
}

int main(int argc, char *argv[]) {
  unsigned char buf[32], msg[32*3];
  int c, single_open=0, retval, len, i;

  eib_device=EIB_DEVICE;

  argv0=argv[0];
  while((c=getopt(argc, argv, "id:ow:csvrhn"))!=EOF) {
    switch(c) {
      
    case 'i':
      device_information();
      return 0;

    case 'd':
      eib_device=optarg;
      break;

    case 'o':
      single_open=1;
      break;

    case 'w':
      write_data_block(optarg);
      return 0;

    case 'c':
      clear_write_buffer();
      return 0;

    case 's':
      reset_device_statistic();
      return 0;

    case 'v':
      fprintf(stdout, "EIBDump %s\n", eibdump_version);
      return 0;

    case 'r':
      reset_bcu();
      return 0;

    case 'n':
      no_write=1;
      break;

    default:
    case '?':
    case 'h':
      fprintf(stdout, "EIBDump version %s, Copyright (C) 2001 Bernd Thallner <bernd@kangaroo.at>\n", eibdump_version);
      fprintf(stdout, "This software comes with ABSOLUTELY NO WARRANTY; for details see COPYING file.\n\n");
      fprintf(stdout, "Usage: %s [options]\nOptions:\n", argv0);
      fprintf(stdout, "  -d <device>       Set EIB device name\n");
      fprintf(stdout, "  -o                Set device to single open\n");
      fprintf(stdout, "  -n                Don't write to BCU, open device O_RDONLY (BCU is not\n");
      fprintf(stdout, "                    automatically set to busmonitor mode and reseted on exit)\n");
      fprintf(stdout, "  -r                Reset BCU and exit\n");
      fprintf(stdout, "  -w <data_block>   Write data block (e.g. \"4c 01 00 60\")\n");
      fprintf(stdout, "  -c                Clear write buffer and exit\n");
      fprintf(stdout, "  -i                Print device information and exit\n");
      fprintf(stdout, "  -s                Reset device statistic and exit\n");
      fprintf(stdout, "  -v                Print program version and exit\n");
      fprintf(stdout, "  -h                Print this help and exit\n\n");
      return 0;
    }
  }

  open_eib();
  if(single_open) {
    if((retval=ioctl(eib_fd, EIB_SINGLE_OPEN))==-1)
      error();
    if(retval>0)
      fprintf(stderr, "%s: Warning: No exclusive access possilbe, device is already %d times opened", argv0, retval);
    else
      dump("set device to single open\n");
  }

  if(!no_write) {
    if(write(eib_fd, "\x46\x01\x00\x60\x90", 5)==-1) // Busmonitor
      error();
    dump("set bcu to busmonitor mode\n");
  }
  signal(SIGINT, cleanup);

  while(1) {
    if((len=read(eib_fd, buf, 32))==-1) {
      write(eib_fd, "\x46\x01\x00\x60\xc0", 5); // Reset
      error();
    }
    if(len) {
      for(i=0; i<len; i++)
	snprintf(&msg[i*3], 4, "%02x ", buf[i]);
      msg[strlen(msg)-1]='\n';
      dump(msg);
    }
  }

  return 0;
}
