/*
     xeibdump.c - XEIBDump Busmonitor for low level EIB driver (BCU 1)

     Copyright (C) 2000 Bernd Thallner <bernd@kangaroo.at>

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include"eib.h"
#include<time.h>
#include<unistd.h>
#include<stdlib.h>
#include<fcntl.h>
#include<stdio.h>
#include<string.h>
#include<errno.h>
#include<sys/ioctl.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<gtk/gtk.h>

#define EIB_DEVICE "/dev/eib"

static char *xeibdump_version="0.2";

GtkWidget *text;
int eib_fd=0, bcu_version;
gint tag;
char *argv0;

int strtoeib(char *in, char *out) {
  int state=0, len=0, tmp;
  do {
    if(state<2) {
      if(*in>='0' && *in<='9')
	tmp=*in-'0';
      else if(*in>='a' && *in<='f')
	tmp=*in-'a'+10;
      else if(*in>='A' && *in<='F')
	tmp=*in-'A'+10;
      else
	return 0;
      if(state)
	out[len]<<=4;
      else
	out[len]=0;
      out[len]+=tmp;
      if(state)
	len++;
      state++;
    } else {
      if(*in!=' ')
	return 0;
      state=0;
    }
  } while(*(++in));
  if(state!=2)
    return 0;
  return len;
}

void dump(char *str, int type) {
  GdkColormap *cmap;
  GdkColor color;
  char time_str[64];
  time_t t;
  struct tm *local_time;

  time(&t);
  local_time=localtime(&t);
  color.red=0;
  color.green=0;
  color.blue=0;
  cmap=gdk_colormap_get_system();

  switch(type) {
  case 1:
    color.blue=0x9000;
    break;
    
  case 2:
    color.red=0x9000;
    break;
  }

  if(!gdk_color_alloc(cmap, &color)) {
    g_error("couldn't allocate color");
  }
  snprintf(time_str, 64, "%02d:%02d:%02d  ", local_time->tm_hour, local_time->tm_min, local_time->tm_sec);
  gtk_text_insert(GTK_TEXT(text), NULL, &color, NULL, time_str, -1);
  gtk_text_insert(GTK_TEXT(text), NULL, &color, NULL, str, -1);
}

void dialog(char *message) {
  GtkWidget *window, *label, *button;

  window=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW (window), "Error");
  gtk_widget_set_usize(window, 200, 100);
  button=gtk_button_new_with_label("OK");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->action_area), button, FALSE, FALSE, 0);
  gtk_widget_show(button);
  label=gtk_label_new(message);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), label, TRUE, TRUE, 0);
  gtk_widget_show(label);
  gtk_widget_show(window);
}

static void clear_window(GtkWidget *w, gpointer data) {
  gtk_text_set_point(GTK_TEXT(text), 0);
  gtk_text_forward_delete(GTK_TEXT(text), gtk_text_get_length(GTK_TEXT(text)));
}

static void single_open(GtkWidget *w, gpointer data) {
  int retval;

  if((retval=ioctl(eib_fd, EIB_SINGLE_OPEN))==-1) {
    dialog(strerror(errno));
    return;
  }
  if(retval>0) {
    char str[64];
    snprintf(str, 64, "No exclusive access possilbe,\ndevice is already %d times opened", retval);
    dialog(str);
  } else
    dump("single open enabled\n", 1);
}

void save_dump_as_ok(GtkWidget *widget, GtkFileSelection *file) {
  char *dump;
  int fd, len;

  dump=gtk_editable_get_chars(GTK_EDITABLE(text), 0, len=gtk_text_get_length(GTK_TEXT(text)));
  if((fd=open(gtk_file_selection_get_filename(GTK_FILE_SELECTION(file)), O_WRONLY | O_CREAT | O_TRUNC , S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))==-1) {
    dialog(strerror(errno));
    return;
  }

  if(write(fd, dump, len)==-1) {
    dialog(strerror(errno));
    return;
  }
  if(close(fd)==-1) {
    dialog(strerror(errno));
    return;
  }
  free(dump);
  gtk_widget_destroy(GTK_WIDGET(file));
}

static void save_dump_as(GtkWidget *w, gpointer data) {
  GtkWidget *file;

  file=gtk_file_selection_new ("Save Dump As");
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION (file)->ok_button), "clicked", (GtkSignalFunc)save_dump_as_ok, file);
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)), "destroy", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file)->cancel_button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(file));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(file), "eib.dump");
  gtk_widget_show(file);
}

void write_data_block_ok(GtkWidget *widget, gpointer entry) {
  char out[32];
  int len;

  if((len=strtoeib(gtk_entry_get_text(GTK_ENTRY(entry)), out))) {
    if(write(eib_fd, out, len)==-1)
      dialog(strerror(errno));
    else {
      char str[256];
      snprintf(str, 256, "write data block: %s\n", gtk_entry_get_text(GTK_ENTRY(entry)));
      dump(str, 1);
    }
  } else
    dialog("Syntax error in string.");
}

void write_data_block(GtkWidget *widget, gpointer data) {
  GtkWidget *window, *label, *hbox, *vbox, *entry, *button;

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(window, 250, 120);
  gtk_window_set_title(GTK_WINDOW (window), "Write Data Block");

  gtk_container_set_border_width(GTK_CONTAINER(window), 10);

  vbox=gtk_vbox_new(TRUE, 10);
  gtk_container_border_width(GTK_CONTAINER(vbox), 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_widget_show(vbox);

  label=gtk_label_new("Data Block: (e.g. \"4c 01 00 60\")");
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);

  entry=gtk_entry_new_with_max_length(32*3-1);
  gtk_box_pack_start(GTK_BOX(vbox), entry, TRUE, TRUE, 0);
  gtk_widget_grab_focus(entry);
  gtk_widget_show(entry);
  
  hbox=gtk_hbox_new(FALSE, 10);
  gtk_box_pack_end(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_widget_show(hbox);

  button=gtk_button_new_with_label("OK");
  gtk_signal_connect(GTK_OBJECT(button), "clicked", (GtkSignalFunc)write_data_block_ok, GTK_OBJECT(entry));
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));

  gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  button=gtk_button_new_with_label("Cancel");
  gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));
  gtk_widget_show(button);

  gtk_widget_show(window);
}

void quit(GtkWidget *widget, gpointer data) {

  if(write(eib_fd, "\x46\x01\x00\x60\xc0", 5)==-1) { // Reset
    perror(argv0);
    gdk_input_remove(tag);
    close(eib_fd);
    gtk_main_quit();
    exit(1);
  }
  gdk_input_remove(tag);
  close(eib_fd);
  gtk_main_quit();
}

void clear_write_buffer(GtkWidget *widget, gpointer data) {

  if(ioctl(eib_fd, EIB_CLEAR_WRITE_BUFFER)==-1)
    dialog(strerror(errno));
  else
    dump("write buffer cleared\n", 2);
}

void device_information(GtkWidget *widget, gpointer data) {
  struct eib_device_information eib_dev_info;
  char tmp[128];
  GtkWidget *window, *text, *button;

  if(ioctl(eib_fd, EIB_GET_DEVICE_INFORMATION, &eib_dev_info)==-1) {
    dialog(strerror(errno));
    return;
  }

  window=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window), "Device Information");
  gtk_widget_set_usize(window, 420, 250);
  button=gtk_button_new_with_label("OK");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->action_area), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  text=gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), FALSE);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  snprintf(tmp, 128, "%s\n", eib_dev_info.string);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "warning_flag=0x%08lx\n", eib_dev_info.warning_flag);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "rb_length=%d\n", eib_dev_info.rb_length);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "wb_length=%d\n", eib_dev_info.wb_length);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "rb_max_length=%d\n", eib_dev_info.rb_max_length);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "wb_max_length=%d\n", eib_dev_info.wb_max_length);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "stat_read_bytes=%ld\n", eib_dev_info.stat_rx_bytes);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "stat_write_bytes=%ld\n", eib_dev_info.stat_tx_bytes);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "stat_read_data_blocks=%ld\n", eib_dev_info.stat_rx_data_blocks);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "stat_write_data_blocks=%ld\n", eib_dev_info.stat_tx_data_blocks);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "stat_max_read_time_data_block=%ld\n", eib_dev_info.stat_max_rx_time_data_block);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);
  snprintf(tmp, 64, "stat_max_write_time_data_block=%ld\n", eib_dev_info.stat_max_tx_time_data_block);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, tmp, -1);

  gtk_widget_show(window);
}

void reset_device_statistic(GtkWidget *widget, gpointer data) {

  if(ioctl(eib_fd, EIB_RESET_DEVICE_STATISTIC)==-1)
    dialog(strerror(errno));
  else
    dump("device statistic reseted\n", 1);
}

void about(GtkWidget *widget, gpointer data) {
  GtkWidget *window, *text, *button;
  GdkFont *font;
  GdkColormap *cmap;
  GdkColor color;
  char str[512];

  window=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(window), "About");
  gtk_widget_set_usize(window, 400, 220);
  button=gtk_button_new_with_label("OK");
  gtk_signal_connect_object(GTK_OBJECT(button), "clicked", (GtkSignalFunc)gtk_widget_destroy, GTK_OBJECT(window));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->action_area), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  text=gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), FALSE);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  cmap=gdk_colormap_get_system();
  color.red=0;
  color.green=0;
  color.blue=0xffff;
  if(!gdk_color_alloc(cmap, &color)) {
    g_error("couldn't allocate color");
  }

  font=gdk_font_load("-*-times-medium-i-normal-*-*-300-*-*-*-*-*-*");
  snprintf(str, 64, "XEIBDump %s\n", xeibdump_version);
  gtk_text_insert(GTK_TEXT(text), font, &color, NULL, str, -1);
  snprintf(str, 512, "\nXEIBDump version %s, Copyright (C) 2001 Bernd Thallner\nBusmonitor for low level EIB driver (BCU 1)\nThis software comes with ABSOLUTELY NO WARRANTY;\nfor details see COPYING file.\n\nAuthor E-Mail: bernd@kangaroo.at\nLinux EIB Web Page: http://under.construction/", xeibdump_version);
  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, str, -1);

  gtk_widget_show(window);
}

void input_callback(gpointer data, gint source, GdkInputCondition condition) {
  int len, i;
  unsigned char buf[32], msg[32*3-1+1];

  if((len=read(eib_fd, buf, 32))==-1) {
    perror(argv0);
    write(eib_fd, "\x46\x01\x00\x60\xc0", 5);
    gdk_input_remove(tag);
    close(eib_fd);
    gtk_main_quit();
    exit(1);
  }
  if(len) {
    for(i=0; i<len; i++)
      snprintf(&msg[i*3], 4, "%02x ", buf[i]);
    msg[strlen(msg)-1]='\n';
    dump(msg, 0);
  }
}

static GtkItemFactoryEntry menu_items[]={
  { "/_File",                        NULL,         NULL,                   0, "<Branch>" },
  { "/File/_Clear Window",           "<control>C", clear_window,           0, NULL },
  { "/File/Single _Open",            "<control>O", single_open,            0, NULL },
  { "/File/_Save Dump As...",        "<control>S", save_dump_as,           0, NULL },
  { "/File/separator",               NULL,         NULL,                   0, "<Separator>" },
  { "/File/Quit",                    "<control>Q", quit,                   0, NULL },
  { "/_Edit",                        NULL,         NULL,                   0, "<Branch>"},
  { "/Edit/_Write Data Block...",    "<control>W", write_data_block,       0, NULL },
  { "/Edit/_Clear Write Buffer",     NULL,         clear_write_buffer,     0, NULL },
  { "/Edit/Device _Information...",  "<control>I", device_information,     0, NULL },
  { "/Edit/_Reset Device Statistic", NULL,         reset_device_statistic, 0, NULL },
  { "/_Help",                        NULL,         NULL,                   0, "<LastBranch>" },
  { "/Help/_About...",               NULL,         about,                  0, NULL } };

int main(int argc, char *argv[]) {
  GtkWidget *window, *scrollbar, *hbox, *vbox, *menubar;
  GtkItemFactory *item_factory;
  GtkAccelGroup *accel_group;
  gint nmenu_items=sizeof(menu_items)/sizeof(menu_items[0]);
  char *eib_device=EIB_DEVICE, c, str[64];

  gtk_init(&argc, &argv);
  argv0=argv[0];
  while((c=getopt(argc, argv, "d:vh"))!=EOF) {
    switch(c) {

    case 'd':
      eib_device=optarg;
      break;

    case 'v':
      fprintf(stdout, "XEIBDump %s\n", xeibdump_version);
      return 0;

    default:
    case '?':
    case 'h':
      fprintf(stdout, "XEIBDump version %s, Copyright (C) 2001 Bernd Thallner <bernd@kangaroo.at>\n", xeibdump_version);
      fprintf(stdout, "This software comes with ABSOLUTELY NO WARRANTY; for details see COPYING file.\n\n");
      fprintf(stdout, "Usage: %s [options]\nOptions:\n", argv0);
      fprintf(stdout, "  -d <device>       Set EIB device name\n");
      fprintf(stdout, "  -v                Print program version number\n");
      fprintf(stdout, "  -h                Print this help\n\n");
      return 0;
    }
  }

  if((eib_fd=open(eib_device, O_RDWR))==-1) {
    snprintf(str, 64, "%s: %s", argv0, eib_device);
    perror(str);
    return 1;
  }

  if((bcu_version=ioctl(eib_fd, EIB_GET_BCU_VERSION))==-1) {
    snprintf(str, 64, "%s: %s", argv0, eib_device);
    perror(str);
    return 1;
  }

  if(bcu_version!=1)
    fprintf(stderr, "%s: %s: warning unsupported bcu version\n", argv0, eib_device);

  if(write(eib_fd, "\x46\x01\x00\x60\x90", 5)==-1) { // Busmonitor
    snprintf(str, 64, "%s: %s", argv0, eib_device);
    perror(str);
    return 1;   
  }

  window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(window, 560, 380);
  gtk_window_set_title(GTK_WINDOW (window), "XEIBDump");
  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
  gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(quit), NULL);

  gtk_container_set_border_width(GTK_CONTAINER(window), 1);

  vbox=gtk_vbox_new(FALSE, 1);
  gtk_container_border_width(GTK_CONTAINER(vbox), 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);
  gtk_widget_show(vbox);

  accel_group=gtk_accel_group_new();
  item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group);
  gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
  menubar=gtk_item_factory_get_widget(item_factory, "<main>");

  gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, TRUE, 0);
  gtk_widget_show(menubar);

  hbox=gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);

  text=gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 0);
  gtk_widget_show(text);

  scrollbar=gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
  gtk_box_pack_start(GTK_BOX(hbox), scrollbar, FALSE, FALSE, 0);
  gtk_widget_show(scrollbar);

  gtk_widget_show(hbox);
  gtk_widget_show(window);

  tag=gdk_input_add(eib_fd, GDK_INPUT_READ, input_callback, NULL);
  dump("eib device opened\n", 1);

  gtk_main ();
  return 0;
}
